/*
 * Wait.c
 *
 * Copyright 2008, Minoru Murashima. All rights reserved.
 * Distributed under the terms of the BSD License.
 *
 *Գס
 *
 *
 *
 *ԸƤ
 */

#include <sys/config.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/buf.h>
#include <machine/interrupt.h>
#include <machine/lock.h>
#include <kern/ProcSignal.h>
#include <kern/timer.h>
#include <kern/Thread.h>
#include <kern/Wait.h>

#include <kern/debug.h>

//#define DEBUG_WAIT 1
#ifdef DEBUG_WAIT
	#define STATIC
	#define INLINE
#else
	#define STATIC	static
	#define INLINE	inline
#endif

//=====================================  ===================================================

/*
 * ȹ¤
 *ԱƶեСѹVmTaskObjΥѹ뤳
 */
typedef struct {
	ThreadObj	*task;
	int			wake;		// ե饰
	int			spinLock;	// ԥå
} Wait;

//===================================== Х륤ݡ =======================================

//===================================== PRIVATE ====================================================

/*
 * Ԥ
 */
STATIC void wait(
	Wait *this,
	const int i_state)
{
	int eflag;

	eflag = enterCli();
	{
		enter_spinlock(&this->spinLock);

		if (this->wake & i_state) {
			this->wake &= ~i_state;
			exit_spinlock(&this->spinLock);
		}
		else {
			delFromSchedule(i_state);
			exit_spinlock(&this->spinLock);
			wait_task();
		}
	}
	exitCli(eflag);
}

/*
 * 򵯤
 *Գ
 * return : YES = NO = Ƥʤ
 */
STATIC int wake(
	Wait *this,
	const int i_state,
	int (*addScheduleMethod)(ThreadObj*, const int))
{
	int ret;
	int eflag;

	eflag = enterCli();
	enter_spinlock(&this->spinLock);
	{
		ret = addScheduleMethod(this->task, i_state);
		if (ret == YES) {
			this->wake &= ~i_state;
		}
		else {
			this->wake |= i_state;
		}
	}
	exit_spinlock(&this->spinLock);
	exitCli(eflag);
	
	return ret;
}

/*
 * ޡϤϥɥ
 *Գ
 */
STATIC void handleWake(
	void *i_task)
{
	addToSchedule(i_task, TASK_SIGNAL_WAIT);
}

/*
 * ǥХԤϥɥ
 */
STATIC void handleDevWake(
	void *i_task)
{
	addToSchedule(i_task, TASK_DEVICE_WAIT);
}

//===================================== PUBLIC =====================================================

/*
 * 󥹥ȥ饯
 */
void waitConstruct(
	WaitObj *i_this,
	ThreadObj *i_task)
{
	Wait *this = (Wait*) i_this;

#ifdef DEBUG
	if (sizeof(Wait) != sizeof(WaitObj)) {
		printk("waitConstruct() object size error! Wait=%d WaitObj=%d\n", sizeof(Wait), sizeof(WaitObj));
		idle();
	}
#endif
	this->task = i_task;
	this->wake = 0;
	this->spinLock = 0;
}

//--------------------------------------------------------------------------------------------------
// Ԥ
//--------------------------------------------------------------------------------------------------

void waitTask()
{
	Wait *this = (Wait*) getWaitTask(getCurrentTask());
	wait(this, TASK_WAIT);
}

/*
 *Գ
 */
void activeTask(
	WaitObj *i_this)
{
	Wait *this = (Wait*) i_this;
	wake(this, TASK_WAIT, addToSchedule);
}

//--------------------------------------------------------------------------------------------------
// Ԥʥ뵯
//--------------------------------------------------------------------------------------------------

void waitTaskSignal()
{
	Wait *this = (Wait*) getWaitTask(getCurrentTask());
	wait(this, TASK_SIGNAL_WAIT);
}

/*
 * ޡ
 */
uint waitTimeSignal(
	const uint i_miliSeconds)
{
	Wait *this = (Wait*) getWaitTask(getCurrentTask());
	void *timer;
	uint remainTime;
	int error;

	if (i_miliSeconds == 0){
		return 0;
	}

	error = setTaskTimer(i_miliSeconds, 0, handleWake, getCurrentTask(), &timer);
	if (error != NOERR){
		return i_miliSeconds;
	}

	wait(this, TASK_SIGNAL_WAIT);

	// Ĥ֤
	remainTime = getRemainTime(timer);

	// 
	freeTimer(timer);

	return remainTime;
}

/*
 *Գ
 */
void activeTaskSignal(
	WaitObj *i_this)
{
	Wait *this = (Wait*) i_this;
	wake(this, TASK_SIGNAL_WAIT, addToSchedule);
}

/*
 * ˵
 *Գ
 */
void activeTaskSoonSignal(
	WaitObj *i_this)
{
	Wait *this = (Wait*) i_this;
	wake(this, TASK_SIGNAL_WAIT, addScheduleSoon);
}

//----------------------------------------------------------
// ǥХԤ
//----------------------------------------------------------

void WaitIntr(
	WaitObj *i_this,
	const uint miliSeconds)
{
	Wait *this = (Wait*) i_this;
	void *timer;
	int error;

	error = setTaskTimer(miliSeconds, 0, handleDevWake, getCurrentTask(), &timer);

	wait(this, TASK_DEVICE_WAIT);

	if (error == NOERR){
		freeTimer(timer);
	}
}

/*
 *Գ
 * return : YES =  NO = ʤ
 */
int activeTaskIntr(
	WaitObj *i_this)
{
	Wait *this = (Wait*) i_this;

	return wake(this, TASK_DEVICE_WAIT, addScheduleSoon);
}

//--------------------------------------------------------------------------------------------------
// ꡼ס
//--------------------------------------------------------------------------------------------------

/*
 * 򥹥꡼פ
 */
void sleepTask(
	const int i_state)		// ץξ֥ե饰
{
	int eflag;

	eflag = enterCli();
	{
		delFromSchedule(i_state);
		wait_task();
	}
	exitCli(eflag);
}

/*
 * 򵯾
 *оݥwaitTask()sleepTask()ԤˤʤäƤ뤳
 */
void wakeTask(
	WaitObj *i_this,
	const int i_state)
{
	Wait *this = (Wait*) i_this;
	addToSchedule(this->task, i_state);
}

/*
 * 򵯾
 *оݥwaitTask()sleepTask()ԤˤʤäƤ뤳
 */
void wakeTaskSoon(
	WaitObj *i_this,
	const int i_state)
{
	Wait *this = (Wait*) i_this;
	addScheduleSoon(this->task, i_state);
}

//--------------------------------------------------------------------------------------------------
// 
//--------------------------------------------------------------------------------------------------

/*
 * λ
 */
int exitSchedule()
{
	cli();
	delFromSchedule(TASK_EXIT);

	return YES;
}

/*
 * ߤ
 *奿åޤǳ߶ػߤ뤳
 * return : task switch YES or NO
 */
int stopSchedule()
{
	cli();
	delFromSchedule(TASK_SIGNAL_STOP);

	return YES;
}

//--------------------------------------------------------------------------------------------------
// ƥॳ
//--------------------------------------------------------------------------------------------------

/*
 *FreeBSD
 * General sleep call.  Suspends the current process until a wakeup is
 * performed on the specified identifier.  The process will then be made
 * runnable with the specified priority.  Sleeps at most timo/hz seconds
 * (0 means no timeout).  If pri includes PCATCH flag, signals are checked
 * before and after sleeping, else signals are not checked.  Returns 0 if
 * awakened, EWOULDBLOCK if the timeout expires.  If PCATCH is set and a
 * signal needs to be delivered, ERESTART is returned if the current system
 * call should be restarted if possible, and EINTR is returned if the system
 * call should be interrupted by the signal (return EINTR).
 */
int tsleep(
	void *chan, 
	int pri, 
	const char *wmesg, 
	int timeout)
{
	struct buf *bp = chan;
	void *timer;
	uint remainTime;
	int error;

	bp->task = getCurrentTask();

	if (0 < timeout) {
		error = setTaskTimer(timeout, 0, handleWake, getCurrentTask(), &timer);
		if (error != NOERR) {
			return 0;
		}

		waitTaskSignal();
		if ((pri & PCATCH) != 0) {
			if (isSigint(TaskGetTaskSignal(getCurrentTask())) == YES) {
				return EINTR;
			}
		}

		// Ĥ֤
		remainTime = getRemainTime(timer);

		// 
		freeTimer(timer);
		
		if (remainTime == 0) {
			return -EWOULDBLOCK;
		}
	}
	else if (timeout == 0) {
		waitTaskSignal();
		if ((pri & PCATCH) != 0) {
			if (isSigint(TaskGetTaskSignal(getCurrentTask())) == YES) {
				return -EINTR;
			}
		}
	}

	return 0;
}

//--------------------------------------------------------------------------------------------------
// ȥȥå
// δ֤ϻȽϥȤϹϥȤ롣
//--------------------------------------------------------------------------------------------------

/*
 * 
 */
void initModifyWaitLock(
	MODIFY_WAIT_LOCK *m_modifyLock)
{
	m_modifyLock->refCount = 0;
	m_modifyLock->modify = 0;
	m_modifyLock->gate = 0;
	AggregateSingleListConstructor(&m_modifyLock->aggregate);
}

/*
 * å
 */
void modifyWaitLock(
	MODIFY_WAIT_LOCK *modifyLock)
{
	for (;;) {
		int eflag = enterCli();
		enter_spinlock(&modifyLock->gate);

		if (0 < modifyLock->refCount) {
			AggregateList *aggregate = &modifyLock->aggregate;
			aggregate->insertEnd(aggregate, getWaitList(getCurrentTask()));
			delFromSchedule(TASK_WAIT);
			exit_spinlock(&modifyLock->gate);
			wait_task();
			exitCli(eflag);
		}
		else {
			modifyLock->modify = YES;
			modifyLock->refCount++;
			exit_spinlock(&modifyLock->gate);
			exitCli(eflag);

			break;
		}
	}
}

/*
 * å
 */
void modifyWaitUnlock(
	MODIFY_WAIT_LOCK *modifyLock)
{
	int eflag;

	ASSERT(0 < modifyLock->refCount);

	eflag = enterCli();
	enter_spinlock(&modifyLock->gate);
	{
		AggregateList *aggregate = &modifyLock->aggregate;
		ThreadObj *waitTask = aggregate->getHead(aggregate);
		modifyLock->modify = NO;
		modifyLock->refCount--;
		if (waitTask != NULL) {
			addToSchedule(waitTask, TASK_WAIT);
		}
	}
	exit_spinlock(&modifyLock->gate);
	exitCli(eflag);
}

/*
 * 
 */
void modifyWaitRef(MODIFY_WAIT_LOCK *modifyLock)
{
	for (;;) {
		int eflag = enterCli();
		enter_spinlock(&modifyLock->gate);

		if (modifyLock->modify == YES) {
			AggregateList *aggregate = &modifyLock->aggregate;
			aggregate->insertEnd(aggregate, getWaitList(getCurrentTask()));
			delFromSchedule(TASK_WAIT);
			exit_spinlock(&modifyLock->gate);
			wait_task();
			exitCli(eflag);
		}
		else {
			modifyLock->refCount++;
			exit_spinlock(&modifyLock->gate);
			exitCli(eflag);

			break;
		}
	}
}

/*
 * Ƚλ
 */
void modifyWaitRefEnd(MODIFY_WAIT_LOCK *modifyLock)
{
	int eflag;

	ASSERT(0 < modifyLock->refCount);

	eflag = enterCli();
	enter_spinlock(&modifyLock->gate);
	{
		AggregateList *aggregate = &modifyLock->aggregate;
		ThreadObj *waitTask = aggregate->getHead(aggregate);
		modifyLock->refCount--;
		if (waitTask != NULL) {
			addToSchedule(waitTask, TASK_WAIT);
		}
	}
	exit_spinlock(&modifyLock->gate);
	exitCli(eflag);
}

//--------------------------------------------------------------------------------------------------
// ƥॳ
//--------------------------------------------------------------------------------------------------

/*
 * return : error number
 */
int sysTsleep(
	const int timeout,		// ॢms
	void **o_waitTask)
{
	if (0 < timeout) {
		void *timer;
		uint remainTime;
		int error = setTaskTimer(timeout, 0, handleWake, getCurrentTask(), &timer);

		if (error != NOERR) {
			return 0;
		}
		
		*o_waitTask = getWaitTask(getCurrentTask());
		waitTaskSignal();

		// Ĥ֤
		remainTime = getRemainTime(timer);

		// 
		freeTimer(timer);
		
		if (remainTime == 0) {
			return -EWOULDBLOCK;
		}
	}
	else if (timeout == 0) {
		*o_waitTask = getWaitTask(getCurrentTask());
		waitTaskSignal();
	}

	return 0;
}

int sysWakeup(
	void *waitTask)
{
	activeTaskSoonSignal(waitTask);
	return 0;
}

int sys_sleep(
	uint seconds)
{
	void *timer;
	uint remainTime;
	int error;

	if (seconds == 0){
		return 0;
	}

	error = setTaskTimer(seconds * 1000, 0, handleWake, getCurrentTask(), &timer);
	if (error != NOERR){
		return seconds;
	}

	sleepTask(TASK_SIGNAL_WAIT);

	// Ĥ֤
	remainTime = getRemainTime(timer);

	// 
	freeTimer(timer);

	return remainTime;
}
