/*	$NetBSD: ata.c,v 1.83 2006/11/16 01:32:47 christos Exp $	*/

/*
 * Copyright (c) 1998, 2001 Manuel Bouyer.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by Manuel Bouyer.
 * 4. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * 2009: modified by minoru murashima.
 */

#include <sys/param.h>
#include <sys/cdefs.h>
#include <sys/errno.h>
#include <sys/systm.h>
#include <sys/queue.h>
#include <sys/event.h>
#include <sys/lock.h>
#include <sys/fcntl.h>
#include <sys/conf.h>
#include <sys/device.h>
#include <sys/ataio.h>
#include <stdlib.h>
#include <machine/intr.h>
#include <GENERIC/locators.h>
#include <ata/atavar.h>

//=====================================  ===================================================

//===================================== Х륤ݡ =======================================

extern struct cfdriver atabus_cd;

//===================================== PRIVATE ====================================================

/*
 * ԥå
 */
static inline int splbio()
{
	// ⤷ʤ
	return 0;
}

/*
 * ԥ󥢥å
 */
static inline void splx(
	int dummy)
{
	// ⤷ʤ
}

struct atabus_initq_head atabus_initq_head = TAILQ_HEAD_INITIALIZER(atabus_initq_head);

/*
 * management of the /dev/atabus* devices
 */
static int atabusopen(
	dev_t dev, 
	int flag, 
	int fmt,
    struct lwp *l)
{
	struct atabus_softc *sc;
	int error, unit = minor(dev);

	if (atabus_cd.cd_ndevs <= unit ||
	    (sc = atabus_cd.cd_devs[unit]) == NULL) {
		return (ENXIO);
	}

	if (sc->sc_flags & ATABUSCF_OPEN) {
		return (EBUSY);
	}

	if ((error = ata_addref(sc->sc_chan)) != 0) {
		return (error);
	}

	sc->sc_flags |= ATABUSCF_OPEN;

	return (0);
}

static int atabusclose(
	dev_t dev, 
	int flag, 
	int fmt,
    struct lwp *l)
{
	struct atabus_softc *sc = atabus_cd.cd_devs[minor(dev)];

	ata_delref(sc->sc_chan);

	sc->sc_flags &= ~ATABUSCF_OPEN;

	return (0);
}

static int atabusioctl(
	dev_t dev, 
	u_long cmd, 
	caddr_t addr, 
	int flag,
    struct lwp *l)
{
	struct atabus_softc *sc = atabus_cd.cd_devs[minor(dev)];
	struct ata_channel *chp = sc->sc_chan;
	int min_drive, max_drive, drive;
	int error;
	int s;

	/*
	 * Enforce write permission for ioctls that change the
	 * state of the bus.  Host adapter specific ioctls must
	 * be checked by the adapter driver.
	 */
	switch (cmd) {
	case ATABUSIOSCAN:
	case ATABUSIODETACH:
	case ATABUSIORESET:
		if ((flag & FWRITE) == 0) {
			return (EBADF);
		}
	}

	switch (cmd) {
	case ATABUSIORESET:
		s = splbio();
		ata_reset_channel(sc->sc_chan, AT_WAIT | AT_POLL);
		splx(s);
		error = 0;
		break;
	case ATABUSIOSCAN:
	{
		if ((chp->ch_drive[0].drive_flags & DRIVE_OLD) ||
		    (chp->ch_drive[1].drive_flags & DRIVE_OLD))
			return (EOPNOTSUPP);
		return (EOPNOTSUPP);
	}
	case ATABUSIODETACH:
	{
		struct atabusioscan_args *a = (struct atabusioscan_args *)addr;
		if ((chp->ch_drive[0].drive_flags & DRIVE_OLD) ||
			(chp->ch_drive[1].drive_flags & DRIVE_OLD)) {
			return (EOPNOTSUPP);
		}
		switch (a->at_dev) {
		case -1:
			min_drive = 0;
			max_drive = 1;
			break;
		case 0:
		case 1:
			min_drive = max_drive = a->at_dev;
			break;
		default:
			return (EINVAL);
		}
		for (drive = min_drive; drive <= max_drive; drive++) {
			if (chp->ch_drive[drive].drv_softc != NULL) {
				error = config_detach(chp->ch_drive[drive].drv_softc, 0);
				if (error) {
					return (error);
				}
				chp->ch_drive[drive].drv_softc = NULL;
			}
		}
		error = 0;
		break;
	}
	default:
		error = ENOTTY;
	}
	return (error);
};

const struct cdevsw atabus_cdevsw = {
	atabusopen, 
	atabusclose, 
	noread, 
	nowrite, 
	atabusioctl,
	nostop, 
	notty, 
	nopoll, 
	nommap, 
	nokqfilter, 
	D_OTHER
};

/*
 * atabus_match:
 *
 *	Autoconfiguration match routine.
 */
static int atabus_match(
	struct device *parent, 
	struct cfdata *cf, 
	void *aux)
{
	struct ata_channel *chp = aux;

	if (chp == NULL) {
		return (0);
	}

	if (cf->cf_loc[ATACF_CHANNEL] != chp->ch_channel &&
		cf->cf_loc[ATACF_CHANNEL] != ATACF_CHANNEL_DEFAULT) {
		return (0);
	}

	return (1);
}

/*
 * atabus_attach:
 *
 *	Autoconfiguration attach routine.
 */
static void atabus_attach(
	struct device *parent, 
	struct device *self, 
	void *aux)
{
	struct atabus_softc *sc = (void *) self;
	struct ata_channel *chp = aux;
	struct atabus_initq *initq;

	sc->sc_chan = chp;

	aprint_normal("\n");
	aprint_naive("\n");

	if (ata_addref(chp)) {
		return;
	}

	initq = malloc(sizeof(*initq));
	initq->atabus_sc = sc;
	TAILQ_INSERT_TAIL(&atabus_initq_head, initq, atabus_initq);
	config_pending_incr();
//	kthread_create(atabus_create_thread, sc);

//	sc->sc_powerhook = powerhook_establish(sc->sc_dev.dv_xname, atabus_powerhook, sc);
//	if (sc->sc_powerhook == NULL) {
//		printf("%s: WARNING: unable to establish power hook\n", sc->sc_dev.dv_xname);
//	}
}

/*
 * atabus_detach:
 *
 *	Autoconfiguration detach routine.
 */
static int atabus_detach(
	struct device *self, 
	int flags)
{
	struct atabus_softc *sc = (void *) self;
	struct ata_channel *chp = sc->sc_chan;
	struct device *dev = NULL;
	int s, i, error = 0;

	/* Shutdown the channel. */
	s = splbio();		/* XXX ALSO NEED AN INTERLOCK HERE. */
	chp->ch_flags |= ATACH_SHUTDOWN;
	splx(s);
	wakeup(&chp->ch_thread);
	while (chp->ch_thread != NULL) {
		(void) tsleep(&chp->ch_flags, PRIBIO, "atadown", 0);
	}

	/* power hook */
//	if (sc->sc_powerhook) {
//		powerhook_disestablish(sc->sc_powerhook);
//	}

	/*
	 * Detach atapibus and its children.
	 */
	if ((dev = chp->atapibus) != NULL) {
		error = config_detach(dev, flags);
		if (error) {
			goto out;
		}
	}

	/*
	 * Detach our other children.
	 */
	for (i = 0; i < chp->ch_ndrive; i++) {
		if (chp->ch_drive[i].drive_flags & DRIVE_ATAPI) {
			continue;
		}
		if ((dev = chp->ch_drive[i].drv_softc) != NULL) {
			error = config_detach(dev, flags);
			if (error) {
				goto out;
			}
		}
	}

 out:
	return (error);
}

/*
 * atabus_activate:
 *
 *	Autoconfiguration activation routine.
 */
static int atabus_activate(
	struct device *self, 
	enum devact act)
{
	struct atabus_softc *sc = (void *) self;
	struct ata_channel *chp = sc->sc_chan;
	struct device *dev = NULL;
	int s, i, error = 0;

	s = splbio();
	switch (act) {
	case DVACT_ACTIVATE:
		error = EOPNOTSUPP;
		break;

	case DVACT_DEACTIVATE:
		/*
		 * We might deactivate the children of atapibus twice
		 * (once bia atapibus, once directly), but since the
		 * generic autoconfiguration code maintains the DVF_ACTIVE
		 * flag, it's safe.
		 */
		if ((dev = chp->atapibus) != NULL) {
			error = config_deactivate(dev);
			if (error)
				goto out;
		}

		for (i = 0; i < chp->ch_ndrive; i++) {
			if ((dev = chp->ch_drive[i].drv_softc) != NULL) {
				error = config_deactivate(dev);
				if (error)
					goto out;
			}
		}
		break;
	}
 out:
	splx(s);

	return (error);
}

//===================================== PUBLIC =====================================================

CFATTACH_DECL(atabus, sizeof(struct atabus_softc), atabus_match, atabus_attach, atabus_detach, atabus_activate);
/*
struct cfattach atabus_ca = { 
	"atabus", 
	{ ((void *)0), ((void *)0) }, 
	sizeof(struct atabus_softc), 
	atabus_match, 
	atabus_attach, 
	atabus_detach, 
	atabus_activate, 
	0, 
	0 
};
*/

/*
 * Start I/O on a controller, for the given channel.
 * The first xfer may be not for our channel if the channel queues
 * are shared.
 *
 * MUST BE CALLED AT splbio()!
 */
void atastart(
	struct ata_channel *chp)
{
	struct atac_softc *atac = chp->ch_atac;
	struct ata_xfer *xfer;

	/* is there a xfer ? */
	if ((xfer = TAILQ_FIRST(&chp->ch_queue->queue_xfer)) == NULL) {
		return;
	}

	/* adjust chp, in case we have a shared queue */
	chp = xfer->c_chp;

	if (chp->ch_queue->active_xfer != NULL) {
		return; /* channel aleady active */
	}
	if (0 < __predict_false(chp->ch_queue->queue_freeze)) {
		if (chp->ch_queue->queue_flags & QF_IDLE_WAIT) {
			chp->ch_queue->queue_flags &= ~QF_IDLE_WAIT;
			wakeup(&chp->ch_queue->queue_flags);
		}
		return; /* queue frozen */
	}
	/*
	 * if someone is waiting for the command to be active, wake it up
	 * and let it process the command
	 */
	if (xfer->c_flags & C_WAITACT) {
		wakeup(xfer);
		return;
	}
	if (atac->atac_claim_hw) {
		if (!(*atac->atac_claim_hw)(chp, 0)) {
			return;
		}
	}

	if (chp->ch_drive[xfer->c_drive].drive_flags & DRIVE_RESET) {
		chp->ch_drive[xfer->c_drive].drive_flags &= ~DRIVE_RESET;
		chp->ch_drive[xfer->c_drive].state = 0;
	}
	chp->ch_queue->active_xfer = xfer;
	TAILQ_REMOVE(&chp->ch_queue->queue_xfer, xfer, c_xferchain);

	if (atac->atac_cap & ATAC_CAP_NOIRQ) {
		KASSERT(xfer->c_flags & C_POLL);
	}

	xfer->c_start(chp, xfer);
}

int ata_addref(
	struct ata_channel *chp)
{
	struct atac_softc *atac = chp->ch_atac;
	struct scsipi_adapter *adapt = &atac->atac_atapi_adapter._generic;
	int s, error = 0;

	s = splbio();
	if (adapt->adapt_refcnt++ == 0 &&
	    adapt->adapt_enable != NULL) {
		error = (*adapt->adapt_enable)(&atac->atac_dev, 1);
		if (error) {
			adapt->adapt_refcnt--;
		}
	}
	splx(s);
	return (error);
}

void ata_delref(
	struct ata_channel *chp)
{
	struct atac_softc *atac = chp->ch_atac;
	struct scsipi_adapter *adapt = &atac->atac_atapi_adapter._generic;
	int s;

	s = splbio();
	if (adapt->adapt_refcnt-- == 1 &&
	    adapt->adapt_enable != NULL) {
		(void) (*adapt->adapt_enable)(&atac->atac_dev, 0);
	}
	splx(s);
}

/*
 * ata_reset_channel:
 *
 *	Reset and ATA channel.
 *
 *	MUST BE CALLED AT splbio()!
 */
void ata_reset_channel(
	struct ata_channel *chp, 
	int flags)
{
	struct atac_softc *atac = chp->ch_atac;
	int drive;

	chp->ch_queue->queue_freeze++;

	/*
	 * If we can poll or wait it's OK, otherwise wake up the
	 * kernel thread to do it for us.
	 */
	if ((flags & (AT_POLL | AT_WAIT)) == 0) {
		if (chp->ch_flags & ATACH_TH_RESET) {
			/* No need to schedule a reset more than one time. */
			chp->ch_queue->queue_freeze--;
			return;
		}
		chp->ch_flags |= ATACH_TH_RESET;
		chp->ch_reset_flags = flags & (AT_RST_EMERG | AT_RST_NOCMD);
		wakeup(&chp->ch_thread);
		return;
	}

	(*atac->atac_bustype_ata->ata_reset_channel)(chp, flags);

	for (drive = 0; drive < chp->ch_ndrive; drive++) {
		chp->ch_drive[drive].state = 0;
	}

	chp->ch_flags &= ~ATACH_TH_RESET;
	if ((flags & AT_RST_EMERG) == 0)  {
		chp->ch_queue->queue_freeze--;
		atastart(chp);
	} 
	else {
		/* make sure that we can use polled commands */
		TAILQ_INIT(&chp->ch_queue->queue_xfer);
		chp->ch_queue->queue_freeze = 0;
		chp->ch_queue->active_xfer = NULL;
	}
}
