/*	$NetBSD: scsi_spc.h,v 1.3 2005/12/11 12:23:50 christos Exp $	*/

/*-
 * Copyright (c) 2005 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Jason R. Thorpe.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the NetBSD
 *	Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * 2009: modified by minoru murashima.
 */

/*
 * SCSI Primary Commands (SPC) --
 *	Commands for all device types
 */

/*
 * Largely written by Julian Elischer (julian@tfs.com)
 * for TRW Financial Systems.
 *
 * TRW Financial Systems, in accordance with their agreement with Carnegie
 * Mellon University, makes this software available to CMU to distribute
 * or use in any manner that they see fit as long as this message is kept with
 * the software. For this reason TFS also grants any other persons or
 * organisations permission to use or modify this software.
 *
 * TFS supplies this software to be publicly redistributed
 * on the understanding that TFS is not responsible for the correct
 * functioning of this software in any circumstances.
 *
 * Ported to run under 386BSD by Julian Elischer (julian@tfs.com) Sept 1992
 */

#ifndef _DEV_SCSIPI_SCSI_SPC_H_
#define	_DEV_SCSIPI_SCSI_SPC_H_

struct scsi_sense_data {
	uint8_t response_code;					/* 1*/
#define	SSD_RCODE(x)		((x) & 0x7f)
#define	SSD_RCODE_CURRENT	0x70
#define	SSD_RCODE_DEFERRED	0x71
#define	SSD_RCODE_VALID		0x80
	uint8_t segment;	/* obsolete */		/* 2*/
	uint8_t flags;							/* 3*/
#define	SSD_SENSE_KEY(x)	((x) & 0x0f)
#define	SSD_ILI				0x20
#define	SSD_EOM				0x40
#define	SSD_FILEMARK		0x80
	uint8_t info[4];						/* 7*/
	uint8_t extra_len;						/* 8*/
	uint8_t csi[4];							/*12*/
	uint8_t asc;							/*13*/
	uint8_t ascq;							/*14*/
	uint8_t fru;							/*15*/
	union {
		uint8_t sks_bytes[3];

		/* ILLEGAL REQUEST */
		struct {
			uint8_t byte0;
#define	SSD_SKS_FP_BIT(x)	((x) & 0x7)
#define	SSD_SKS_FP_BPV		0x08
#define	SSK_SKS_FP_CD		0x40	/* 1=command, 0=data */
			uint8_t val[2];
		} field_pointer;

		/* RECOVERED ERROR, HARDWARE ERROR, MEDIUM ERROR */
		struct {
			uint8_t byte0;
			uint8_t val[2];
		} actual_retry_count;

		/* NOT READY, NO SENSE */
		struct {
			uint8_t byte0;
			uint8_t val[2];
		} progress_indication;

		/* COPY ABORTED */
		struct {
			uint8_t byte0;
#define	SSD_SKS_SP_BIT(x)	((x) & 0x7)
#define	SSD_SKS_SP_BPV		0x08
#define	SSD_SKS_SP_SD		0x20	/* 0=param list, 1=segment desc */
			uint8_t val[2];
		} segment_pointer;
	} sks;									/*18*/
#define	SSD_SKSV		0x80		/* byte0 of sks field */
	uint8_t extra_bytes[14];				/*32*/ /* really variable length */
};

#endif /* _DEV_SCSIPI_SCSI_SPC_H_ */
