/*	$NetBSD: systm.h,v 1.228 2008/09/23 22:20:24 pooka Exp $	*/

/*-
 * Copyright (c) 1982, 1988, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 * (c) UNIX System Laboratories, Inc.
 * All or some portions of this file are derived from material licensed
 * to the University of California by American Telephone and Telegraph
 * Co. or Unix System Laboratories, Inc. and are reproduced herein with
 * the permission of UNIX System Laboratories, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)systm.h	8.7 (Berkeley) 3/29/95
 *
 * 2008: modified by minoru murashima.
 */

#ifndef _SYS_SYSTM_H_
#define _SYS_SYSTM_H_

#include <sys/lock.h>
#include <sys/uio.h>
#include <sys/kauth.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#define	KM_SLEEP	0x00000001	/* can sleep */
#define	KM_NOSLEEP	0x00000002	/* don't sleep */

#define PWR_RESUME		0
#define PWR_SUSPEND		1
#define PWR_STANDBY		2
#define PWR_SOFTRESUME	3
#define PWR_SOFTSUSPEND	4
#define PWR_SOFTSTANDBY	5

typedef unsigned int km_flag_t;

struct lwp {
	struct kauth_cred *l_cred;	/* Cached credentials */
};

extern int lbolt;		/* once a second sleep address */

/*
 * General function declarations.
 */
extern int	nullop(void *);
extern int	enodev(void);
extern int	enosys(void);
extern int	enoioctl(void);
extern int	enxio(void);
extern int	eopnotsupp(void);

/* Compatibility with old, non-interlocked tsleep call */
#define	tsleep(chan, pri, wmesg, timo)					\
	ltsleep(chan, pri, wmesg, timo, NULL)

static inline void panic(
	const char *str, ...)
{
	va_list ap;

	va_start(ap, str);
	vfprintf(stdout, str, ap);
	va_end(ap);

	exit(0);
}

static inline void aprint_normal(
	const char *str, ...)
{
	va_list ap;

	va_start(ap, str);
//	vfprintf(stdout, str, ap);
	va_end(ap);
}

static inline void aprint_naive(
	const char *str, ...)
{
	// ⤷ʤ
}

static inline void aprint_error(
	const char *str, ...)
{
	va_list ap;

	va_start(ap, str);
	vfprintf(stdout, str, ap);
	va_end(ap);
}

static inline void *kmem_alloc(
	size_t size, 
	km_flag_t dummy)
{
	return malloc(size);
}

static inline void kmem_free(
	void *mem, 
	size_t dummy)
{
	free(mem);
}

static inline int copyin(const void *from, void *to, size_t len)
{
	memcpy(to, from, len);
	return 0;
}

static inline int copyout(const void *from, void *to, size_t len)
{
	memcpy(to, from, len);
	return 0;
}

static inline int copyinstr(const void *from, void *to, size_t len, size_t *plen)
{
	memcpy(to, from, len);
	return 0;
}

static inline int subyte(
	void *uaddr, 
	char x)
{
	 *(char*) uaddr = x;
}

extern void	cpu_configure(void);
extern int	ltsleep(volatile const void *, int, const char *, int, volatile struct simplelock *);
extern void	wakeup(volatile const void *);
extern int	uiomove(void *, size_t, struct uio *);
extern int	scanc (u_int, const u_char *, const u_char *, int);
extern int	seltrue(dev_t, int, struct lwp *);

/*
 * Power management hooks.
 */
extern void	*powerhook_establish(const char *, void (*)(int, void *), void *);

#ifdef DEBUG
 #define KASSERT(__cond) if (!(__cond)) {panic("Assertion! : %s line=%d \n",__FILE__, __LINE__);}
#else
 #define KASSERT(__cond)
#endif

#endif	/* !_SYS_SYSTM_H_ */
